import os
from typing import (  # noqa: F401
    Any,
    List,
    Optional,
    Set,
    TypeVar,
    Union,
    cast)
from typing_extensions import Type

from ruamel import yaml
from ruamel.yaml.error import StreamMark

from yatiml.exceptions import RecognitionError, SeasoningError
from yatiml.introspection import defaulted_attributes
from yatiml.irecognizer import IRecognizer
from yatiml.util import ScalarType, scalar_type_to_tag

_Any = TypeVar('_Any')


class Node:
    """A wrapper class for yaml Nodes that provides utility functions.

    This class defines a number of helper function for you to use \
    when writing _yatiml_sweeten() and _yatiml_savorize() functions. It \
    also gives access to the underlying `yaml.Node`, so you can do \
    anything ruamel.yaml can do if you're willing to dive into its \
    internals.

    Attributes:
        yaml_node: The wrapped yaml Node. You can read and modify this,\
                yourself if you want, or replace it completely with a \
                new yaml.Node.
    """

    def __init__(self, node: yaml.Node) -> None:
        """Create a Node for a particular node.

        The member functions will act on the contained node.

        Args:
            node: The node to provide help for.
        """
        self.yaml_node = node

    def __str__(self) -> str:
        """Convert to a human-readable string."""
        return 'Node({})'.format(self.yaml_node)

    def is_scalar(self, typ: Union[Type, None] = _Any) -> bool:
        """Returns True iff this represents a scalar node.

        If a type is given, checks that the ScalarNode represents this \
        type. Type may be `str`, `int`, `float`, `bool`, or `None`.

        If no type is given, any ScalarNode will return True.
        """
        if isinstance(self.yaml_node, yaml.ScalarNode):
            if typ != _Any and typ in scalar_type_to_tag:
                if typ is None:
                    typ = type(None)
                return cast(str, self.yaml_node.tag) == scalar_type_to_tag[typ]

            if typ is _Any:
                return True
            raise ValueError('Invalid scalar type passed to is_scalar()')
        return False

    def is_mapping(self) -> bool:
        """Returns True iff this represents a mapping node."""
        return isinstance(self.yaml_node, yaml.MappingNode)

    def is_sequence(self) -> bool:
        """Returns True iff this represents a sequence node."""
        return isinstance(self.yaml_node, yaml.SequenceNode)

    # Functions for Scalar nodes
    def get_value(self) -> ScalarType:
        """Returns the value of a Scalar node.

        Use is_scalar(type) to check which type the node has.
        """
        if self.yaml_node.tag == 'tag:yaml.org,2002:str':
            return str(self.yaml_node.value)
        if self.yaml_node.tag == 'tag:yaml.org,2002:int':
            return int(self.yaml_node.value)
        if self.yaml_node.tag == 'tag:yaml.org,2002:float':
            return float(self.yaml_node.value)
        if self.yaml_node.tag == 'tag:yaml.org,2002:bool':
            return self.yaml_node.value in ['TRUE', 'True', 'true']
        if self.yaml_node.tag == 'tag:yaml.org,2002:null':
            return None
        raise RuntimeError('This node with tag {} is not of the right type'
                           ' for get_value()'.format(self.yaml_node.tag))

    def set_value(self, value: ScalarType) -> None:
        """Sets the value of the node to a scalar value.

        After this, is_scalar(type(value)) will return true.

        Args:
            value: The value to set this node to, a str, int, float, \
                    bool, or None.
        """
        if isinstance(value, bool):
            value_str = 'true' if value else 'false'
        else:
            value_str = str(value)
        start_mark = self.yaml_node.start_mark
        end_mark = self.yaml_node.end_mark
        # If we're of a class type, then we want to keep that tag so that the
        # correct Constructor is called. If we're a built-in type, set the tag
        # to the appropriate YAML tag.
        tag = self.yaml_node.tag
        if tag.startswith('tag:yaml.org,2002:'):
            tag = scalar_type_to_tag[type(value)]
        new_node = yaml.ScalarNode(tag, value_str, start_mark, end_mark)
        self.yaml_node = new_node

    # Functions for Mapping nodes

    def make_mapping(self) -> None:
        """Replaces the node with a new, empty mapping.

        Note that this will work on the Node object that is passed to \
        a _yatiml_savorize() or _yatiml_sweeten() function, but not on \
        any of its attributes or items. If you need to set an attribute \
        to a complex value, build a yaml.Node representing it and use \
        set_attribute with that.
        """
        start_mark = StreamMark('generated node', 0, 0, 0)
        end_mark = StreamMark('generated node', 0, 0, 0)
        self.yaml_node = yaml.MappingNode('tag:yaml.org,2002:map', list(),
                                          start_mark, end_mark)

    def has_attribute(self, attribute: str) -> bool:
        """Whether the node has an attribute with the given name.

        Use only if is_mapping() returns True.

        Args:
            attribute: The name of the attribute to check for.

        Returns:
            True iff the attribute is present.
        """
        return any([
            key_node.value == attribute for key_node, _ in self.yaml_node.value
        ])

    def has_attribute_type(self, attribute: str, typ: Optional[Type]) -> bool:
        """Whether the given attribute exists and has a compatible type.

        Returns true iff the attribute exists and is an instance of \
        the given type. Matching between types passed as typ and \
        yaml node types is as follows:

        +---------+-------------------------------------------+
        |   typ   |                 yaml                      |
        +=========+===========================================+
        |   str   |      ScalarNode containing string         |
        +---------+-------------------------------------------+
        |   int   |      ScalarNode containing int            |
        +---------+-------------------------------------------+
        |  float  |      ScalarNode containing float          |
        +---------+-------------------------------------------+
        |   bool  |      ScalarNode containing bool           |
        +---------+-------------------------------------------+
        |   None  |      ScalarNode containing null           |
        +---------+-------------------------------------------+
        |   list  |      SequenceNode                         |
        +---------+-------------------------------------------+
        |   dict  |      MappingNode                          |
        +---------+-------------------------------------------+

        Args:
            attribute: The name of the attribute to check.
            typ: The type to check against.

        Returns:
            True iff the attribute exists and matches the type.
        """
        if not self.has_attribute(attribute):
            return False

        attr_node = self.get_attribute(attribute).yaml_node

        if typ in scalar_type_to_tag:
            tag = scalar_type_to_tag[typ]
            return cast(str, attr_node.tag) == tag
        elif typ == list:
            return isinstance(attr_node, yaml.SequenceNode)
        elif typ == dict:
            return isinstance(attr_node, yaml.MappingNode)

        raise ValueError('Invalid argument for typ attribute')

    def get_attribute(self, attribute: str) -> 'Node':
        """Returns the node representing the given attribute's value.

        Use only if is_mapping() returns true.

        Args:
            attribute: The name of the attribute to retrieve.

        Raises:
            KeyError: If the attribute does not exist.

        Returns:
            A node representing the value.
        """
        matches = [
            value_node for key_node, value_node in self.yaml_node.value
            if key_node.value == attribute
        ]
        if len(matches) != 1:
            raise SeasoningError(
                'Attribute not found, or found multiple times: {}'.format(
                    matches))
        return Node(matches[0])

    def set_attribute(self, attribute: str,
                      value: Union[ScalarType, yaml.Node]) -> None:
        """Sets the attribute to the given value.

        Use only if is_mapping() returns True.

        If the attribute does not exist, this adds a new attribute, \
        if it does, it will be overwritten.

        If value is a str, int, float, bool or None, the attribute will \
        be set to this value. If you want to set the value to a list or \
        dict containing other values, build a yaml.Node and pass it here.

        Args:
            attribute: Name of the attribute whose value to change.
            value: The value to set.
        """
        start_mark = StreamMark('generated node', 0, 0, 0)
        end_mark = StreamMark('generated node', 0, 0, 0)
        if isinstance(value, str):
            value_node = yaml.ScalarNode(
                    'tag:yaml.org,2002:str', value,
                    start_mark, end_mark)  # type: yaml.Node
        elif isinstance(value, bool):
            value_str = 'true' if value else 'false'
            value_node = yaml.ScalarNode('tag:yaml.org,2002:bool', value_str,
                                         start_mark, end_mark)
        elif isinstance(value, int):
            value_node = yaml.ScalarNode('tag:yaml.org,2002:int', str(value),
                                         start_mark, end_mark)
        elif isinstance(value, float):
            value_node = yaml.ScalarNode('tag:yaml.org,2002:float', str(value),
                                         start_mark, end_mark)
        elif value is None:
            value_node = yaml.ScalarNode('tag:yaml.org,2002:null', '',
                                         start_mark, end_mark)
        elif isinstance(value, yaml.Node):
            value_node = value
        else:
            raise TypeError('Invalid kind of value passed to set_attribute()')

        attr_index = self.__attr_index(attribute)
        if attr_index is not None:
            key_node = self.yaml_node.value[attr_index][0]
            self.yaml_node.value[attr_index] = key_node, value_node
        else:
            key_node = yaml.ScalarNode('tag:yaml.org,2002:str', attribute,
                                       start_mark, end_mark)
            self.yaml_node.value.append((key_node, value_node))
            if isinstance(self.yaml_node, yaml.MappingNode):
                self.yaml_node.flow_style = False

    def remove_attribute(self, attribute: str) -> None:
        """Remove an attribute from the node.

        Use only if is_mapping() returns True.

        Args:
            attribute: The name of the attribute to remove.
        """
        attr_index = self.__attr_index(attribute)
        if attr_index is not None:
            self.yaml_node.value.pop(attr_index)

    def remove_attributes_with_default_values(self, cls: Type) -> None:
        """Remove attributes with default values.

        If you have a class with many optional attributes, then saving
        it to YAML may yield a very large dictionary with many values
        set to e.g. None. If there's no risk of creating an ambiguity,
        then you may want to remove any attributes whose value matches
        the default.

        This function can be used in _yatiml_sweeten() to do that. For
        `cls`, pass the `cls` first argument of _yatiml_sweeten().

        If the default for the parameter is not the same as the default
        of the attribute, for example in this common situation:

        .. code-block:: python

            def __init__(
                    self, my_list: Optional[List[int]] = None) -> None:
                if my_list is None:
                    my_list = list()
                self.my_list = my_list


        then this function will not remove the attribute if it is an
        empty list, because it compares with `None` in the type
        annotation. To fix that, you can define `_yatiml_defaults` like
        this:

        .. code-block:: python

            def __init__(
                    self, my_list: Optional[List[int]] = None) -> None:
                if my_list is None:
                    my_list = list()
                self.my_list = my_list

            _yatiml_defaults = {'my_list', []}  # type: Dict[str, Any]

            @classmethod
            def _yatiml_sweeten(cls, node: yatiml.Node) -> None:
                node.remove_attributes_with_default_values(cls)


        Note that this function currently only works for the built-in
        types bool, float, int, str and for None values, not for
        classes or enums.

        Use only if is_mapping() returns True.

        Args:
            cls: The class we're sweetening.
        """
        def matches(value_node: yaml.Node, default: Any) -> bool:
            if value_node.tag == 'tag:yaml.org,2002:null':
                return default is None

            if value_node.tag == 'tag:yaml.org,2002:int':
                return int(value_node.value) == int(default)

            if value_node.tag == 'tag:yaml.org,2002:float':
                return float(value_node.value) == float(default)

            if value_node.tag == 'tag:yaml.org,2002:bool':
                if default is False:
                    return (
                            str(value_node.value).lower() == 'n' or
                            str(value_node.value).lower() == 'no' or
                            str(value_node.value).lower() == 'false' or
                            str(value_node.value).lower() == 'off')
                elif default is True:
                    return (
                            str(value_node.value).lower() == 'y' or
                            str(value_node.value).lower() == 'yes' or
                            str(value_node.value).lower() == 'true' or
                            str(value_node.value).lower() == 'on')
                return False

            return bool(value_node.value == default)

        defaults = defaulted_attributes(cls)

        self.yaml_node.value = [
                (name_node, value_node)
                for name_node, value_node in self.yaml_node.value
                if (
                    name_node.value not in defaults or
                    not matches(value_node, defaults[name_node.value]))]

    def rename_attribute(self, attribute: str, new_name: str) -> None:
        """Renames an attribute.

        Use only if is_mapping() returns true.

        If the attribute does not exist, this will do nothing.

        Args:
            attribute: The (old) name of the attribute to rename.
            new_name: The new name to rename it to.
        """
        for key_node, _ in self.yaml_node.value:
            if key_node.value == attribute:
                key_node.value = new_name
                break

    def unders_to_dashes_in_keys(self) -> None:
        """Replaces underscores with dashes in key names.

        For each attribute in a mapping, this replaces any underscores \
        in its keys with dashes. Handy because Python does not \
        accept dashes in identifiers, while some YAML-based formats use \
        dashes in their keys.
        """
        for key_node, _ in self.yaml_node.value:
            key_node.value = key_node.value.replace('_', '-')

    def dashes_to_unders_in_keys(self) -> None:
        """Replaces dashes with underscores in key names.

        For each attribute in a mapping, this replaces any dashes in \
        its keys with underscores. Handy because Python does not \
        accept dashes in identifiers, while some YAML-based file \
        formats use dashes in their keys.
        """
        for key_node, _ in self.yaml_node.value:
            key_node.value = key_node.value.replace('-', '_')

    def seq_attribute_to_map(self,
                             attribute: str,
                             key_attribute: str,
                             value_attribute: Optional[str] = None,
                             strict: Optional[bool] = True) -> None:
        """Converts a sequence attribute to a map.

        This function takes an attribute of this Node that is \
        a sequence of mappings and turns it into a mapping of mappings. \
        It assumes that each of the mappings in the original sequence \
        has an attribute containing a unique value, which it will use \
        as a key for the new outer mapping.

        An example probably helps. If you have a Node representing \
        this piece of YAML::

            items:
            - item_id: item1
              description: Basic widget
              price: 100.0
            - item_id: item2
              description: Premium quality widget
              price: 200.0

        and call seq_attribute_to_map('items', 'item_id'), then the \
        Node will be modified to represent this::

            items:
              item1:
                description: Basic widget
                price: 100.0
              item2:
                description: Premium quality widget
                price: 200.0

        which is often more intuitive for people to read and write.

        If the attribute does not exist, or is not a sequence of \
        mappings, this function will silently do nothing. If the keys \
        are not unique and strict is False, it will also do nothing. If \
        the keys are not unique and strict is True, it will raise an \
        error.

        With thanks to the makers of the Common Workflow Language for \
        the idea.

        Args:
            attribute: Name of the attribute whose value to modify.
            key_attribute: Name of the attribute in each item to use \
                    as a key for the new mapping.
            value_attribute: Name of the attribute in each item to use \
                    for the value in the new mapping, if only a key and \
                    value have been given.
            strict: Whether to give an error if the intended keys are \
                    not unique.

        Raises:
            yatiml.SeasoningError: If the keys are not unique and \
                    strict is True.
        """
        if not self.has_attribute(attribute):
            return

        attr_node = self.get_attribute(attribute)
        if not attr_node.is_sequence():
            return

        start_mark = attr_node.yaml_node.start_mark
        end_mark = attr_node.yaml_node.end_mark

        # check that all list items are mappings and that the keys are unique
        # strings
        seen_keys = set()  # type: Set[str]
        for item in attr_node.seq_items():
            key_attr_node = item.get_attribute(key_attribute)
            if not key_attr_node.is_scalar(str):
                raise SeasoningError(
                    ('Attribute names must be strings in'
                     'YAtiML, {} is not a string.').format(key_attr_node))
            if key_attr_node.get_value() in seen_keys:
                if strict:
                    raise SeasoningError(
                        ('Found a duplicate key {}: {} when'
                         ' converting from sequence to mapping'.format(
                             key_attribute, key_attr_node.get_value())))
                return
            seen_keys.add(key_attr_node.get_value())  # type: ignore

        # construct mapping
        mapping_values = list()
        for item in attr_node.seq_items():
            # we've already checked that it's a SequenceNode above
            key_node = item.get_attribute(key_attribute).yaml_node
            item.remove_attribute(key_attribute)
            if value_attribute is not None:
                value_node = item.get_attribute(value_attribute).yaml_node
                if len(item.yaml_node.value) == 1:
                    # no other attributes, use short form
                    mapping_values.append((key_node, value_node))
                else:
                    mapping_values.append((key_node, item.yaml_node))
            else:
                mapping_values.append((key_node, item.yaml_node))

        # create mapping node
        mapping = yaml.MappingNode('tag:yaml.org,2002:map', mapping_values,
                                   start_mark, end_mark)
        self.set_attribute(attribute, mapping)

    def map_attribute_to_seq(self,
                             attribute: str,
                             key_attribute: str,
                             value_attribute: Optional[str] = None) -> None:
        """Converts a mapping attribute to a sequence.

        This function takes an attribute of this Node whose value \
        is a mapping or a mapping of mappings and turns it into a \
        sequence of mappings. Each entry in the original mapping is \
        converted to an entry in the list. If only a key attribute is \
        given, then each entry in the original mapping must map to a \
        (sub)mapping. This submapping becomes the corresponding list \
        entry, with the key added to it as an additional attribute. If a \
        value attribute is also given, then an entry in the original \
        mapping may map to any object. If the mapped-to object is a \
        mapping, the conversion is as before, otherwise a new \
        submapping is created, and key and value are added using the \
        given key and value attribute names.

        An example probably helps. If you have a Node representing \
        this piece of YAML::

            items:
              item1:
                description: Basic widget
                price: 100.0
              item2:
                description: Premium quality widget
                price: 200.0

        and call map_attribute_to_seq('items', 'item_id'), then the \
        Node will be modified to represent this::

            items:
            - item_id: item1
              description: Basic widget
              price: 100.0
            - item_id: item2
              description: Premium quality widget
              price: 200.0

        which once converted to an object is often easier to deal with \
        in code.

        Slightly more complicated, this YAML::

            items:
              item1: Basic widget
              item2:
                description: Premium quality widget
                price: 200.0

        when passed through map_attribute_to_seq('items', 'item_id', \
        'description'), will result in th equivalent of::

            items:
            - item_id: item1
              description: Basic widget
            - item_id: item2
              description: Premium quality widget
              price: 200.0

        If the attribute does not exist, or is not a mapping, this \
        function will silently do nothing.

        With thanks to the makers of the Common Workflow Language for \
        the idea.

        Args:
            attribute: Name of the attribute whose value to modify.
            key_attribute: Name of the new attribute in each item to \
                    add with the value of the key.
            value_attribute: Name of the new attribute in each item to \
                    add with the value of the key.
        """
        if not self.has_attribute(attribute):
            return

        attr_node = self.get_attribute(attribute)
        if not attr_node.is_mapping():
            return

        start_mark = attr_node.yaml_node.start_mark
        end_mark = attr_node.yaml_node.end_mark
        object_list = []
        for item_key, item_value in attr_node.yaml_node.value:
            item_value_node = Node(item_value)
            if not item_value_node.is_mapping():
                if value_attribute is None:
                    return
                ynode = item_value_node.yaml_node
                item_value_node.make_mapping()
                item_value_node.set_attribute(value_attribute, ynode)

            item_value_node.set_attribute(key_attribute, item_key.value)
            object_list.append(item_value_node.yaml_node)
        seq_node = yaml.SequenceNode('tag:yaml.org,2002:seq', object_list,
                                     start_mark, end_mark)
        self.set_attribute(attribute, seq_node)

    # Functions for sequences

    def seq_items(self) -> List['Node']:
        """Returns the items in the sequence.

        Use only if is_sequence() returns True.
        """
        return list(map(Node, self.yaml_node.value))

    # Private helpers

    def __attr_index(self, attribute: str) -> Optional[int]:
        """Finds an attribute's index in the yaml_node.value list."""
        attr_index = None
        for i, (key_node, _) in enumerate(self.yaml_node.value):
            if key_node.value == attribute:
                attr_index = i
                break
        return attr_index


class UnknownNode:
    """Utility functions for recognizing nodes.

    This class defines a number of helper function for you to use \
    when writing _yatiml_recognize() functions.

    Attributes:
        yaml_node: The yaml.Node wrapped by this object.
    """

    def __init__(self, recognizer: IRecognizer, node: yaml.Node) -> None:
        """Create an UnknownNode for a particular mapping node.

        The member functions will act on the contained node.

        Args:
            node: The node to operate on.
        """
        self.__recognizer = recognizer
        self.yaml_node = node

    def __str__(self) -> str:
        """Convert to a human-readable string."""
        return 'UnknownNode({})'.format(self.yaml_node)

    def require_scalar(self, *args: Type) -> None:
        """Require the node to be a scalar.

        If additional arguments are passed, these are taken as a list \
        of valid types; if the node matches one of these, then it is \
        accepted.

        Example:
            # Match either an int or a string
            node.require_scalar(int, str)

        Arguments:
            args: One or more types to match one of.
        """
        node = Node(self.yaml_node)
        if len(args) == 0:
            if not node.is_scalar():
                raise RecognitionError(('{}{}A scalar is required').format(
                    self.yaml_node.start_mark, os.linesep))
        else:
            for typ in args:
                if node.is_scalar(typ):
                    return
            raise RecognitionError(
                ('{}{}A scalar of type {} is required').format(
                    self.yaml_node.start_mark, os.linesep, args))

    def require_mapping(self) -> None:
        """Require the node to be a mapping."""
        if not isinstance(self.yaml_node, yaml.MappingNode):
            raise RecognitionError(('{}{}A mapping is required here').format(
                self.yaml_node.start_mark, os.linesep))

    def require_sequence(self) -> None:
        """Require the node to be a sequence."""
        if not isinstance(self.yaml_node, yaml.SequenceNode):
            raise RecognitionError(('{}{}A sequence is required here').format(
                self.yaml_node.start_mark, os.linesep))

    def require_attribute(
            self, attribute: str, typ: Union[None, Type] = _Any) -> None:
        """Require an attribute on the node to exist.

        This implies that the node must be a mapping.

        If `typ` is given, the attribute must have this type.

        Args:
            attribute: The name of the attribute / mapping key.
            typ: The type the attribute must have.
        """
        self.require_mapping()
        attr_nodes = [
            value_node for key_node, value_node in self.yaml_node.value
            if key_node.value == attribute
        ]
        if len(attr_nodes) == 0:
            raise RecognitionError(
                ('{}{}Missing required attribute {}').format(
                    self.yaml_node.start_mark, os.linesep, attribute))
        attr_node = attr_nodes[0]

        if typ != _Any:
            recognized_types, message = self.__recognizer.recognize(
                attr_node, cast(Type, typ))
            if len(recognized_types) == 0:
                raise RecognitionError(message)

    def require_attribute_value(
            self, attribute: str,
            value: Union[int, str, float, bool, None]) -> None:
        """Require an attribute on the node to have a particular value.

        This requires the attribute to exist, and to have the given value \
        and corresponding type. Handy for in your _yatiml_recognize() \
        function.

        Args:
            attribute: The name of the attribute / mapping key.
            value: The value the attribute must have to recognize an \
                    object of this type.

        Raises:
            yatiml.RecognitionError: If the attribute does not exist, \
                    or does not have the required value.
        """
        found = False
        for key_node, value_node in self.yaml_node.value:
            if (key_node.tag == 'tag:yaml.org,2002:str'
                    and key_node.value == attribute):
                found = True
                node = Node(value_node)
                if not node.is_scalar(type(value)):
                    raise RecognitionError(
                            ('{}{}Incorrect attribute type where value {}'
                             ' of type {} was required').format(
                                self.yaml_node.start_mark, os.linesep,
                                value, type(value)))
                if node.get_value() != value:
                    raise RecognitionError(
                        ('{}{}Incorrect attribute value'
                         ' {} where {} was required').format(
                             self.yaml_node.start_mark, os.linesep,
                             value_node.value, value))

        if not found:
            raise RecognitionError(
                ('{}{}Required attribute {} not found').format(
                    self.yaml_node.start_mark, os.linesep, attribute))

    def require_attribute_value_not(
            self, attribute: str,
            value: Union[int, str, float, bool, None]) -> None:
        """Require an attribute on the node to not have a given value.

        This requires the attribute to exist, and to not have the given
        value.

        Args:
            attribute: The name of the attribute / mapping key.
            value: The value the attribute must not have to recognize an
                    object of this type.

        Raises:
            yatiml.RecognitionError: If the attribute does not exist,
                    or has the required value.
        """
        found = False
        for key_node, value_node in self.yaml_node.value:
            if (key_node.tag == 'tag:yaml.org,2002:str'
                    and key_node.value == attribute):
                found = True
                node = Node(value_node)
                if not node.is_scalar(type(value)):
                    return
                if node.get_value() == value:
                    raise RecognitionError(
                        ('{}{}Incorrect attribute value'
                         ' {} where {} was not allowed').format(
                             self.yaml_node.start_mark, os.linesep,
                             value_node.value, value))

        if not found:
            raise RecognitionError(
                ('{}{}Required attribute {} not found').format(
                    self.yaml_node.start_mark, os.linesep, attribute))



import inspect
from typing import Any, Dict, Generator, Tuple
from typing_extensions import Type


def class_subobjects(
        class_: Type) -> Generator[Tuple[str, Type, bool], None, None]:
    """Find the aggregated subobjects of an object.

    These are the public attributes.

    Args:
        class_: The class whose subobjects to return.

    Yields:
        Tuples (name, type, required) describing subobjects.
    """
    argspec = inspect.getfullargspec(class_.__init__)
    defaults = argspec.defaults if argspec.defaults else []
    num_optional = len(defaults)
    first_optional = len(argspec.args) - num_optional

    for i, attr_name in enumerate(argspec.args):
        if attr_name == 'self':
            continue
        if attr_name == '_yatiml_extra':
            continue
        attr_type = argspec.annotations.get(attr_name, Any)
        yield attr_name, attr_type, i < first_optional


def defaulted_attributes(class_: Type) -> Dict[str, Any]:
    """Returns attributes with defaulted values.

    Args:
        class_: The class to inspect.

    Returns:
        A dictionary containing attribute names and their default
        values.
    """
    argspec = inspect.getfullargspec(class_.__init__)
    defaults = argspec.defaults if argspec.defaults else []
    num_optional = len(defaults)
    first_optional = len(argspec.args) - num_optional

    if hasattr(class_, '_yatiml_defaults'):
        user_defaults = class_._yatiml_defaults     # type: Dict[str, Any]
    else:
        user_defaults = {}

    print('c: {}, ud: {}'.format(class_, user_defaults))

    result = dict()     # type: Dict[str, Any]
    for i, default in enumerate(defaults):
        arg_name = argspec.args[first_optional + i]
        if arg_name in user_defaults:
            default = user_defaults[arg_name]
        result[arg_name] = default
    return result



import inspect
import logging
from typing import Any, cast
from typing_extensions import TYPE_CHECKING, Type

from ruamel import yaml

from yatiml.helpers import Node

if TYPE_CHECKING:
    from yatiml.dumper import Dumper  # noqa: F401

logger = logging.getLogger(__name__)


class Representer:
    """A yaml Representer class for user-defined types.

    For ruamel.yaml to dump a class correctly, it needs a representer \
    function for that class. YAtiML provides this generic representer \
    which represents classes based on their public attributes by \
    default, with an optional user override using a member function.
    """

    def __init__(self, class_: Type) -> None:
        """Creates a new Representer for the given class.

        Args:
            class_: The class to represent.
        """
        self.class_ = class_

    def __call__(self, dumper: 'Dumper', data: Any) -> yaml.Node:
        """Represents the class as a MappingNode.

        Args:
            dumper: The dumper to use.
            data: The user-defined object to dump.

        Returns:
            A yaml.Node representing the object.
        """
        # make a dict with attributes
        logger.info('Representing {} of class {}'.format(
            data, self.class_.__name__))
        if hasattr(data, '_yatiml_attributes'):
            logger.debug('Found _yatiml_attributes()')
            attributes = data._yatiml_attributes()
            if attributes is None:
                raise RuntimeError(('{}._yatiml_attributes() returned None,'
                                    ' where a dict was expected.').format(
                                        self.class_.__name__))
        else:
            logger.debug(
                'No _yatiml_attributes() found, using public attributes')
            argspec = inspect.getfullargspec(data.__init__)
            attribute_names = list(argspec.args[1:])
            attrs = [(name, getattr(data, name)) for name in attribute_names
                     if name != '_yatiml_extra']
            if '_yatiml_extra' in attribute_names:
                if not hasattr(data, '_yatiml_extra'):
                    raise RuntimeError(
                        ('Class {} takes _yatiml_extra but has '
                         ' no _yatiml_extra attribute, and no '
                         ' _yatiml_attributes().').format(
                             self.class_.__name__))
                attrs.extend(data._yatiml_extra.items())
            attributes = yaml.comments.CommentedMap(attrs)

        # convert to a yaml.MappingNode
        represented = dumper.represent_mapping('tag:yaml.org,2002:map',
                                               attributes)  # type: yaml.Node

        # sweeten
        cnode = Node(represented)
        self.__sweeten(dumper, self.class_, cnode)
        # __sweeten() checks this, so can cast safely
        represented = cast(yaml.Node, cnode.yaml_node)

        logger.debug('End representing {}'.format(data))
        return represented

    def __sweeten(self, dumper: 'Dumper', class_: Type, node: Node) -> None:
        """Applies the user's _yatiml_sweeten() function(s), if any.

        Sweetening is done for the base classes first, then for the \
        derived classes, down the hierarchy to the class we're \
        constructing.

        Args:
            dumper: The dumper that is dumping this object.
            class_: The type of the object to be dumped.
            represented_object: The object to be dumped.
        """
        for base_class in class_.__bases__:
            if base_class in dumper.yaml_representers:
                logger.debug('Sweetening for class {}'.format(
                    self.class_.__name__))
                self.__sweeten(dumper, base_class, node)
        if hasattr(class_, '_yatiml_sweeten'):
            class_._yatiml_sweeten(node)
            if not isinstance(node.yaml_node, yaml.Node):
                raise RuntimeError(
                        ('After sweetening an object of class {},'
                         ' node.yaml_node is not a yaml.Node. Please'
                         ' check your _yatiml_sweeten() function.'
                         ).format(class_.__name__))


class EnumRepresenter:
    """A yaml Representer class for user-defined enum types.

    For ruamel.yaml to dump a class correctly, it needs a representer \
    function for that class. YAtiML provides this generic representer \
    which represents enum classes based on the names of their values by \
    default, with an optional user override using a member function.
    """

    def __init__(self, class_: Type) -> None:
        """Creates a new Representer for the given class.

        Args:
            class_: The class to represent.
        """
        self.class_ = class_

    def __call__(self, dumper: 'Dumper', data: Any) -> yaml.Node:
        """Represents the class as a ScalarNode.

        Args:
            dumper: The dumper to use.
            data: The user-defined object to dump.

        Returns:
            A yaml.Node representing the object.
        """
        # make a ScalarNode of type str with name of value
        logger.info('Representing {} of class {}'.format(
            data, self.class_.__name__))

        # convert to a yaml.ScalarNode
        start_mark = yaml.error.StreamMark('Generated node', 0, 0, 0)
        end_mark = start_mark
        represented = yaml.ScalarNode(
                'tag:yaml.org,2002:str',
                data.name, start_mark, end_mark)    # type: yaml.Node

        # sweeten
        snode = Node(represented)
        if hasattr(self.class_, '_yatiml_sweeten'):
            self.class_._yatiml_sweeten(snode)
            represented = snode.yaml_node

        logger.debug('End representing {}'.format(data))
        return represented


class UserStringRepresenter:
    """A yaml Representer class for user-defined string types.

    For ruamel.yaml to dump a class correctly, it needs a representer \
    function for that class. YAtiML provides this generic representer \
    which represents user-defined string classes as strings.
    """

    def __init__(self, class_: Type) -> None:
        """Creates a new Representer for the given class.

        Args:
            class_: The class to represent.
        """
        self.class_ = class_

    def __call__(self, dumper: 'Dumper', data: Any) -> yaml.Node:
        """Represents the class as a Node.

        Args:
            dumper: The dumper to use.
            data: The user-defined object to dump.

        Returns:
            A yaml.Node representing the object.
        """
        # make a ScalarNode of type str with name of value
        logger.info('Representing {} of class {}'.format(
            data, self.class_.__name__))

        # convert to a yaml.ScalarNode
        represented = dumper.represent_str(str(data))   # type: yaml.Node

        # sweeten
        snode = Node(represented)
        if hasattr(self.class_, '_yatiml_sweeten'):
            self.class_._yatiml_sweeten(snode)
            if not isinstance(snode.yaml_node, yaml.Node):
                raise RuntimeError(
                        ('After sweetening an object of class {},'
                         ' node.yaml_node is not a yaml.Node. Please'
                         ' check your _yatiml_sweeten() function.'
                         ).format(self.class_.__name__))
            represented = snode.yaml_node

        logger.debug('End representing {}'.format(data))
        return represented



import enum
import logging
import os
from collections import UserString
from datetime import datetime
from textwrap import indent
from typing import Dict, List
from typing_extensions import Type

from ruamel import yaml

from yatiml.exceptions import RecognitionError
from yatiml.helpers import Node, UnknownNode
from yatiml.introspection import class_subobjects
from yatiml.irecognizer import IRecognizer, RecResult
from yatiml.util import (
        generic_type_args, is_generic_mapping, is_generic_sequence,
        is_generic_union, scalar_type_to_tag, type_to_desc, bool_union_fix)

logger = logging.getLogger(__name__)


class Recognizer(IRecognizer):
    """Functions for recognizing objects as types."""

    def __init__(self, registered_classes: Dict[str, Type]) -> None:
        """Create a Recognizer.

        Args:
            registered_classes: The registered tags and corresponding \
                    classes.
        """
        self.__registered_classes = registered_classes

    def __recognize_scalar(self, node: yaml.Node,
                           expected_type: Type) -> RecResult:
        """Recognize a node that we expect to be a scalar.

        Args:
            node: The node to recognize.
            expected_type: The type it is expected to be.

        Returns:
            A list of recognized types and an error message
        """
        logger.debug('Recognizing as a scalar')
        if (isinstance(node, yaml.ScalarNode)
                and node.tag == scalar_type_to_tag[expected_type]):
            return [expected_type], ''
        message = 'Failed to recognize a {}\n{}\n'.format(
            type_to_desc(expected_type), node.start_mark)
        return [], message

    def __recognize_list(self, node: yaml.Node,
                         expected_type: Type) -> RecResult:
        """Recognize a node that we expect to be a list of some kind.

        Args:
            node: The node to recognize.
            expected_type: List[...something...]

        Returns
            expected_type and the empty string if it was recognized,
                    [] and an error message otherwise.
        """
        logger.debug('Recognizing as a list')
        if not isinstance(node, yaml.SequenceNode):
            message = '{}{}Expected a list here.'.format(
                node.start_mark, os.linesep)
            return [], message
        item_type = generic_type_args(expected_type)[0]
        for item in node.value:
            recognized_types, message = self.recognize(item, item_type)
            if len(recognized_types) == 0:
                return [], message
            if len(recognized_types) > 1:
                recognized_types = [
                    List[t]  # type: ignore
                    for t in recognized_types
                ]
                return recognized_types, message

        return [expected_type], ''

    def __recognize_dict(self, node: yaml.Node,
                         expected_type: Type) -> RecResult:
        """Recognize a node that we expect to be a dict of some kind.

        Args:
            node: The node to recognize.
            expected_type: Dict[str, ...something...]

        Returns:
            expected_type if it was recognized, [] otherwise.
        """
        logger.debug('Recognizing as a dict')
        if not issubclass(generic_type_args(expected_type)[0], str):
            raise RuntimeError(
                'YAtiML only supports dicts with strings as keys')
        if not isinstance(node, yaml.MappingNode):
            message = '{}{}Expected a dict/mapping here'.format(
                node.start_mark, os.linesep)
            return [], message
        value_type = generic_type_args(expected_type)[1]
        for _, value in node.value:
            recognized_value_types, message = self.recognize(value, value_type)
            if len(recognized_value_types) == 0:
                return [], message
            if len(recognized_value_types) > 1:
                return [
                    Dict[str, t]  # type: ignore
                    for t in recognized_value_types
                ], message  # type: ignore

        return [expected_type], ''

    def __recognize_union(self, node: yaml.Node,
                          expected_type: Type) -> RecResult:
        """Recognize a node that we expect to be one of a union of types.

        Args:
            node: The node to recognize.
            expected_type: Union[...something...]

        Returns:
            The specific type that was recognized, multiple, or none.
        """
        logger.debug('Recognizing as a union')
        recognized_types = []
        message = ''
        union_types = generic_type_args(expected_type)
        logger.debug('Union types {}'.format(union_types))
        for possible_type in union_types:
            recognized_type, msg = self.recognize(node, possible_type)
            if len(recognized_type) == 0:
                message += msg
            recognized_types.extend(recognized_type)
        recognized_types = list(set(recognized_types))
        if bool in recognized_types and bool_union_fix in recognized_types:
            recognized_types.remove(bool_union_fix)

        if len(recognized_types) == 0:
            return recognized_types, message
        elif len(recognized_types) > 1:
            message = ('{}{}Could not determine which of the following types'
                       ' this is: {}').format(node.start_mark, os.linesep,
                                              recognized_types)
            return recognized_types, message

        return recognized_types, ''

    def __recognize_user_class(self, node: yaml.Node,
                               expected_type: Type) -> RecResult:
        """Recognize a user-defined class in the node.

        This tries to recognize only exactly the specified class. It \
        recurses down into the class's attributes, but not to its \
        subclasses. See also __recognize_user_classes().

        Args:
            node: The node to recognize.
            expected_type: A user-defined class.

        Returns:
            A list containing the user-defined class, or an empty list.
        """
        logger.debug('Recognizing as a user-defined class')
        loc_str = '{}{}'.format(node.start_mark, os.linesep)
        if hasattr(expected_type, '_yatiml_recognize'):
            try:
                unode = UnknownNode(self, node)
                expected_type._yatiml_recognize(unode)
                return [expected_type], ''
            except RecognitionError as e:
                if len(e.args) > 0:
                    message = ('Error recognizing a {}\n{}because of the'
                               ' following error(s): {}').format(
                                   expected_type.__class__, loc_str,
                                   indent(e.args[0], '    '))
                else:
                    message = 'Error recognizing a {}\n{}'.format(
                        expected_type.__class__, loc_str)
                return [], message
        else:
            if issubclass(expected_type, enum.Enum):
                if (not isinstance(node, yaml.ScalarNode)
                        or node.tag != 'tag:yaml.org,2002:str'):
                    message = 'Expected an enum value from {}\n{}'.format(
                        expected_type.__class__, loc_str)
                    return [], message
            elif (issubclass(expected_type, UserString)
                  or issubclass(expected_type, str)):
                if (not isinstance(node, yaml.ScalarNode)
                        or node.tag != 'tag:yaml.org,2002:str'):
                    message = 'Expected a string matching {}\n{}'.format(
                        expected_type.__class__, loc_str)
                    return [], message
            else:
                # auto-recognize based on constructor signature
                if not isinstance(node, yaml.MappingNode):
                    message = 'Expected a dict/mapping here\n{}'.format(
                        loc_str)
                    return [], message

                for attr_name, type_, required in class_subobjects(
                        expected_type):
                    cnode = Node(node)
                    # try exact match first, dashes if that doesn't match
                    for name in [attr_name, attr_name.replace('_', '-')]:
                        if cnode.has_attribute(name):
                            subnode = cnode.get_attribute(name)
                            recognized_types, message = self.recognize(
                                subnode.yaml_node, type_)
                            if len(recognized_types) == 0:
                                message = ('Failed when checking attribute'
                                           ' {}:\n{}').format(
                                               name, indent(message, '    '))
                                return [], message
                            break
                    else:
                        if required:
                            message = (
                                'Error recognizing a {}\n{}because it'
                                ' is missing an attribute named {}').format(
                                    expected_type.__name__, loc_str, attr_name)
                            if '_' in attr_name:
                                message += ' or maybe {}.\n'.format(
                                    attr_name.replace('_', '-'))
                            else:
                                message += '.\n'
                            return [], message

            return [expected_type], ''

    def __recognize_user_classes(self, node: yaml.Node,
                                 expected_type: Type) -> RecResult:
        """Recognize a user-defined class in the node.

        This returns a list of classes from the inheritance hierarchy \
        headed by expected_type which match the given node and which \
        do not have a registered derived class that matches the given \
        node. So, the returned classes are the most derived matching \
        classes that inherit from expected_type.

        This function recurses down the user's inheritance hierarchy.

        Args:
            node: The node to recognize.
            expected_type: A user-defined class.

        Returns:
            A list containing matched user-defined classes.
        """
        # Let the user override with an explicit tag
        if node.tag in self.__registered_classes:
            return [self.__registered_classes[node.tag]], ''

        recognized_subclasses = []
        message = ''
        for other_class in self.__registered_classes.values():
            if expected_type in other_class.__bases__:
                sub_subclasses, msg = self.__recognize_user_classes(
                    node, other_class)
                recognized_subclasses.extend(sub_subclasses)
                if len(sub_subclasses) == 0:
                    message += msg

        if len(recognized_subclasses) == 0:
            recognized_subclasses, msg = self.__recognize_user_class(
                node, expected_type)
            if len(recognized_subclasses) == 0:
                message += msg

        if len(recognized_subclasses) == 0:
            message = ('Failed to recognize a {}\n{}\nbecause of the following'
                       ' error(s):\n{}').format(expected_type.__name__,
                                                node.start_mark,
                                                indent(msg, '    '))
            return [], message

        if len(recognized_subclasses) > 1:
            message = ('{}{} Could not determine which of the following types'
                       ' this is: {}').format(node.start_mark, os.linesep,
                                              recognized_subclasses)
            return recognized_subclasses, message

        return recognized_subclasses, ''

    def recognize(self, node: yaml.Node, expected_type: Type) -> RecResult:
        """Figure out how to interpret this node.

        This is not quite a type check. This function makes a list of \
        all types that match the expected type and also the node, and \
        returns that list. The goal here is not to test validity, but \
        to determine how to process this node further.

        That said, it will recognize built-in types only in case of \
        an exact match.

        Args:
            node: The YAML node to recognize.
            expected_type: The type we expect this node to be, based \
                    on the context provided by our type definitions.

        Returns:
            A list of matching types.
        """
        logger.debug('Recognizing {} as a {}'.format(node, expected_type))
        recognized_types = None
        if expected_type in [
                str, int, float, bool, bool_union_fix, datetime, None,
                type(None)
        ]:
            recognized_types, message = self.__recognize_scalar(
                node, expected_type)
        elif is_generic_union(expected_type):
            recognized_types, message = self.__recognize_union(
                node, expected_type)
        elif is_generic_sequence(expected_type):
            recognized_types, message = self.__recognize_list(
                node, expected_type)
        elif is_generic_mapping(expected_type):
            recognized_types, message = self.__recognize_dict(
                node, expected_type)
        elif expected_type in self.__registered_classes.values():
            recognized_types, message = self.__recognize_user_classes(
                node, expected_type)

        if recognized_types is None:
            raise RecognitionError(
                ('Could not recognize for type {},'
                 ' is it registered?').format(expected_type))
        logger.debug('Recognized types {} matching {}'.format(
            recognized_types, expected_type))
        return recognized_types, message



from collections import UserString
import enum
import json
import logging
from pathlib import Path
from typing import Any, AnyStr, Callable, IO, List, Optional, Union, cast
from typing_extensions import Type

from ruamel import yaml
from ruamel.yaml.events import (
        AliasEvent, DocumentEndEvent, MappingEndEvent, MappingStartEvent,
        ScalarEvent, SequenceEndEvent, SequenceStartEvent)

from yatiml.representers import (EnumRepresenter, Representer,
                                 UserStringRepresenter)

logger = logging.getLogger(__name__)


class JsonDumperState(enum.Enum):
    NONE = 0,
    SEQUENCE = 1,
    SEQUENCE_FIRST = 2,
    MAPPING_KEY = 3,
    MAPPING_KEY_FIRST = 4,
    MAPPING_VALUE = 5


class Dumper(yaml.RoundTripDumper):
    """The YAtiML Dumper class.

    Derive your own Dumper class from this one, then add classes to it \
    using add_to_dumper(). Set output_format = "json" to output JSON.
    """
    output_format = 'yaml'

    def __init__(
            self, stream: Any, default_style: Any,
            default_flow_style: Optional[bool], canonical: Optional[int],
            indent: Optional[int], width: Optional[int],
            allow_unicode: Optional[bool], line_break: Any, encoding: Any,
            explicit_start: Optional[bool], explicit_end: Optional[bool],
            version: Any, tags: Any, block_seq_indent: Any,
            top_level_colon_align: Any, prefix_colon: Any
            ) -> None:
        """Create a JsonDumper, called by ruamel.yaml."""
        yaml.RoundTripDumper.__init__(
                self, stream, default_style, default_flow_style, canonical,
                indent, width, allow_unicode, line_break, encoding,
                explicit_start, explicit_end, version, tags, block_seq_indent,
                top_level_colon_align, prefix_colon)

        self._json_state = [JsonDumperState.NONE]
        self._cur_indent = 0

        if self.requested_indent is not None:
            self._kv_sep = ': '
        else:
            self._kv_sep = ':'

    def emit(self, event: yaml.events.Event) -> None:
        """Emit an event.

        This is called by ruamel.yaml, and we dispatch to the original
        implementation (which emits YAML) or our emit_json() below if
        the user selected JSON output.

        Args:
            event: The event to output.
        """
        if self.output_format == 'json':
            self.emit_json(event)
        else:
            yaml.RoundTripDumper.emit(self, event)

    def emit_json(self, event: yaml.events.Event) -> None:
        """Emit a YAML event to the JSON output.

        Args:
            event: The event to output.
        """
        if isinstance(event, AliasEvent):
            raise RuntimeError('Aliases are not supported by JSON')
        elif isinstance(event, SequenceEndEvent):
            self._cur_indent -= self.best_sequence_indent
            self._do_endline()
            self.stream.write(']')
            self._json_state.pop()
        elif isinstance(event, MappingEndEvent):
            self._cur_indent -= self.best_map_indent
            self._do_endline()
            self.stream.write('}')
            self._json_state.pop()
        elif isinstance(event, DocumentEndEvent):
            self._do_endline()
        else:
            # top-level scalar, or continuing a sequence or mapping
            cur_level = len(self._json_state) - 1
            cur_state = self._json_state[cur_level]

            # write separator if necessary
            if cur_state == JsonDumperState.SEQUENCE:
                self.stream.write(',')
                self._do_endline()
            elif cur_state == JsonDumperState.MAPPING_KEY:
                self.stream.write(',')
                self._do_endline()
            elif cur_state == JsonDumperState.MAPPING_VALUE:
                self.stream.write(self._kv_sep)

            # output value
            if isinstance(event, SequenceStartEvent):
                self.stream.write('[')
                self._cur_indent += self.best_sequence_indent
                self._do_endline()
                self._json_state.append(JsonDumperState.SEQUENCE_FIRST)
            elif isinstance(event, MappingStartEvent):
                self.stream.write('{')
                self._cur_indent += self.best_map_indent
                self._do_endline()
                self._json_state.append(JsonDumperState.MAPPING_KEY_FIRST)
            elif isinstance(event, ScalarEvent):
                if event.tag == 'tag:yaml.org,2002:str':
                    self.stream.write(json.dumps(
                        event.value, ensure_ascii=not self.allow_unicode))
                elif event.tag == 'tag:yaml.org,2002:null':
                    self.stream.write('null')
                elif event.tag == 'tag:yaml.org,2002:bool':
                    self.stream.write(event.value.lower())
                elif event.tag == 'tag:yaml.org,2002:timestamp':
                    self.stream.write(json.dumps(
                        event.value, ensure_ascii=not self.allow_unicode))
                else:
                    self.stream.write(event.value)

            # set next state
            if cur_state == JsonDumperState.SEQUENCE_FIRST:
                self._json_state[cur_level] = JsonDumperState.SEQUENCE
            elif cur_state == JsonDumperState.MAPPING_KEY_FIRST:
                self._json_state[cur_level] = JsonDumperState.MAPPING_VALUE
            elif cur_state == JsonDumperState.MAPPING_KEY:
                self._json_state[cur_level] = JsonDumperState.MAPPING_VALUE
            elif cur_state == JsonDumperState.MAPPING_VALUE:
                self._json_state[cur_level] = JsonDumperState.MAPPING_KEY

    def _do_endline(self) -> None:
        """Write EOL and subsequent indent if enabled."""
        if self.requested_indent is not None:
            self.stream.write(self.best_line_break)
            self.stream.write(' ' * self._cur_indent)


# Python errors if we define classes as Union[List[Type], Type]
# So List[Type] it is, and if the user ignores that and passes
# a single class, it'll work anyway, with a little mypy override.
def add_to_dumper(dumper: Type, classes: List[Type]) -> None:
    """Register user-defined classes with the Dumper.

    This enables the Dumper to write objects of your classes to a \
    YAML file. Note that all the arguments are types, not instances!

    Args:
        dumper: Your dumper class(!), derived from yatiml.Dumper
        classes: One or more classes to add.
    """
    if not isinstance(classes, list):
        classes = [classes]  # type: ignore
    for class_ in classes:
        if issubclass(class_, enum.Enum):
            dumper.add_representer(class_, EnumRepresenter(class_))
        elif issubclass(class_, str) or issubclass(class_, UserString):
            dumper.add_representer(class_, UserStringRepresenter(class_))
        else:
            dumper.add_representer(class_, Representer(class_))


def dumps_function(*args: Type) -> Callable[[Any], str]:
    """Create a dumps function for the given types.

    This function returns a callable object which takes an object
    and returns a string containing the YAML serialisation of that
    object. The type of the object, and any other custom types
    needed, must have been passed to dumps_function().

    Note that only custom classes need to be passed, the built-in
    types are, well, built in.

    Examples:

        .. code-block:: python

          dumps = yatiml.dumps_function()
          yaml_text = dumps({'x': 1})
          assert yaml_text == 'x: 1\\n'

        .. code-block:: python

          dumps_config = yatiml.dumps_function(Config, Setting)
          yaml_text = dumps_config(my_config)

        Here, Config is the top-level class, and Setting is
        another class that is used by Config somewhere. Note that any
        object can be passed to the resulting dumps function, as long
        as it is of a built-in type or its type and any types needed
        to represent it have been passed to dumps_function().

    Args:
        *args: Any custom types needed.

    Returns:
        A function that takes an object and produces a YAML string
        representing it.
    """
    class UserDumper(Dumper):
        pass

    add_to_dumper(UserDumper, list(args))

    class DumpsFunction:
        """Dumps objects to a YAML string."""
        def __init__(self, dumper: Type[Dumper]) -> None:
            """Create a dumps function."""
            self.dumper = dumper

        def __call__(self, obj: Any) -> str:
            """Dump the object to a YAML string.

            Args:
                obj: An object to dump.

            Returns:
                A string containing its YAML representation.
            """
            return cast(str, yaml.dump(obj, Dumper=self.dumper))

    return DumpsFunction(UserDumper)


def dump_function(
        *args: Type) -> Callable[[Any, Union[str, Path, IO[AnyStr]]], None]:
    """Create a dump function for the given types.

    This function returns a callable object which takes an object
    and a stream, Path or file name, and writes the YAML serialisation
    of that object to the target. The type of the object, and any other
    custom types needed, must have been passed to dump_function().

    Note that only custom classes should be passed, the built-in
    types are, well, built in.

    Examples:

        .. code-block:: python

          dump = yatiml.dump_function()
          with open('test.yaml', 'w') as f:
              dump({'x': 1}, f)
          # will write 'x: 1\\n' to test.yaml

        .. code-block:: python

          my_config = Config(...)
          dump_config = yatiml.dump_function(Config, Setting)
          dump_config(my_config, 'config.yaml')

        Here, Config is the top-level class, and Setting is
        another class that is used by Config somewhere. Note that any
        object can be passed to the resulting dump function, as long
        as it is of a built-in type or its type and any types needed
        to represent it have been passed to dump_function().

    Args:
        *args: Any custom types needed.

    Returns:
        A function that takes an object and a stream, file name or Path,
        and writes a YAML string representing the object to the file.
    """
    class UserDumper(Dumper):
        pass

    add_to_dumper(UserDumper, list(args))

    class DumpFunction:
        """Dumps objects to a stream."""
        def __init__(self, dumper: Type[Dumper]) -> None:
            """Create a dumps function."""
            self.dumper = dumper

        def __call__(
                self, obj: Any, sink: Union[str, Path, IO[AnyStr]]) -> None:
            """Dump the object to a file or stream.

            The sink argument may be a string containing a path, a
            pathlib.Path object containing a path, or a stream to write
            to directly (e.g. an open file).

            Args:
                obj: An object to dump.
                sink: A place to save it to.
            """
            if isinstance(sink, str):
                sink = Path(sink)

            if isinstance(sink, Path):
                with sink.open('w') as f:
                    yaml.dump(obj, f, Dumper=UserDumper)
            else:
                yaml.dump(obj, sink, Dumper=UserDumper)

    return DumpFunction(UserDumper)


def dumps_json_function(*args: Type) -> Callable[..., str]:
    """Create a dumps function for the given types that writes JSON.

    This function returns a callable object which takes an object
    and returns a string containing the JSON serialisation of that
    object. The type of the object, and any other custom types
    needed, must have been passed to dumps_json_function().

    By default, the produced JSON is in ASCII and in a compact format
    without newlines or spaces. To make it more readable, you can
    pass an `indent` keyword argument to the returned function, which
    causes the output to be formatted with the given indent.

    To output unicode characters as-is, pass `ensure_ascii=False` to
    the returned dumps function.

    Note that only custom classes need to be passed, the built-in
    types are, well, built in.

    Examples:

        .. code-block:: python

          dumps_json = yatiml.dumps_json_function()
          json_text = dumps_json({'x': 1})
          assert json_text == '{"x": 1}'

          json_text = dumps_json({'x': 1}, indent=2)
          assert json_text == (
              '{\\n'
              '  "x": 1\\n'
              '}\\n')


        .. code-block:: python

          dumps_config_to_json = yatiml.dumps_json_function(Config, Setting)
          json_text = dumps_config_to_json(my_config)

        Here, Config is the top-level class, and Setting is
        another class that is used by Config somewhere. Note that any
        object can be passed to the resulting dumps function, as long
        as it is of a built-in type or its type and any types needed
        to represent it have been passed to dumps_json_function().

    Args:
        *args: Any custom types needed.

    Returns:
        A function that takes an object and produces a YAML string
        representing it.
    """
    class UserDumper(Dumper):
        output_format = 'json'

    add_to_dumper(UserDumper, list(args))

    class DumpsJsonFunction:
        """Dumps objects to a JSON string."""
        def __init__(self, dumper: Type[Dumper]) -> None:
            """Create a dumps function."""
            self.dumper = dumper

        def __call__(
                self, obj: Any, *,
                indent: Optional[int] = None, ensure_ascii: bool = True
                ) -> str:
            """Dump the object to a JSON string.

            Args:
                obj: An object to dump.

            Returns:
                A string containing its JSON representation.
            """
            return cast(str, yaml.dump(
                obj, Dumper=self.dumper,
                indent=indent, allow_unicode=not ensure_ascii))

    return DumpsJsonFunction(UserDumper)


def dump_json_function(
        *args: Type) -> Callable[..., None]:
    """Create a dump function for the given types that writes JSON.

    This function returns a callable object which takes an object
    and a stream, Path or file name, and writes the JSON serialisation
    of that object to the target. The type of the object, and any other
    custom types needed, must have been passed to dump_json_function().

    By default, the produced JSON is in ASCII and in a compact format
    without newlines or spaces. To make it more readable, you can
    pass an `indent` keyword argument to the returned function, which
    causes the output to be formatted with the given indent.

    To output unicode characters as-is, pass `ensure_ascii=False` to
    the returned dumps function.

    Note that only custom classes need to be passed, the built-in
    types are, well, built in.

    Examples:

        .. code-block:: python

          dump_json = yatiml.dump_json_function()
          with open('test.json', 'w') as f:
              dump_json({'x': 1}, f)
          # will write '{"x": 1}' to test.json

          dump_json({'x': 1}, 'test.json')
          # will write the same, but nicely formatted

        .. code-block:: python

          my_config = Config(...)
          dump_config_as_json = yatiml.dump_function(Config, Setting)
          dump_config_as_json(my_config, 'config.json', indent=2)

        Here, Config is the top-level class, and Setting is
        another class that is used by Config somewhere. Note that any
        object can be passed to the resulting dump function, as long
        as it is of a built-in type or its type and any types needed
        to represent it have been passed to dump_json_function().

    Args:
        *args: Any custom types needed.

    Returns:
        A function that takes an object and a stream, file name or Path,
        and writes a JSON string representing the object to the file.
    """
    class UserDumper(Dumper):
        output_format = 'json'

    add_to_dumper(UserDumper, list(args))

    class DumpJsonFunction:
        """Dumps objects to a stream."""
        def __init__(self, dumper: Type[Dumper]) -> None:
            """Create a dump function."""
            self.dumper = dumper

        def __call__(
                self, obj: Any, sink: Union[str, Path, IO[AnyStr]],
                indent: Optional[int] = None, ensure_ascii: bool = True
                ) -> None:
            """Dump the object to a file or stream as JSON.

            The sink argument may be a string containing a path, a
            pathlib.Path object containing a path, or a stream to write
            to directly (e.g. an open file).

            Args:
                obj: An object to dump.
                sink: A place to save it to.
            """
            if isinstance(sink, str):
                sink = Path(sink)

            if isinstance(sink, Path):
                with sink.open('w') as f:
                    yaml.dump(
                            obj, f, Dumper=UserDumper,
                            indent=indent, allow_unicode=not ensure_ascii)
            else:
                yaml.dump(
                        obj, sink, Dumper=UserDumper,
                        indent=indent, allow_unicode=not ensure_ascii)

    return DumpJsonFunction(UserDumper)
